@@ -31,24 +31,21 @@
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 """
 import getopt
+import json
 import sys
 import os
 import tempfile
+import traceback
 import collections
-
+import glob
 from . import utils
-
 from . import passes
-
 from . import interface
-
 from . import provenance
-
 from . import pool
-
 from . import driver
-
 from . import config
+# from . import rop_guided_dce
 
 instructions = """slash has three modes of use:
 
@@ -58,7 +55,7 @@ instructions = """slash has three modes of use:
 
     slash [options] <manifest>
 
-    Previrtualize a compilation unit based on its manifest.
+    Specialize a compilation unit based on its manifest.
 
         --help                     : Print this.
 
@@ -74,23 +71,23 @@ instructions = """slash has three modes of use:
         --debug-pass=<tag>         : Debug opt's pass (<tag> should be the debug pragma string of the pass)
         --debug                    : Pass the debug flag into all calls to opt (too much information usually)
         --print-after-all          : Pass the print-after-all flag into all calls to opt
-        --devirt=<type>            : Devirtualize indirect function calls 
-                                     (<type> should be either none, dsa, sea_dsa or cha_dsa)
-        --intra-spec-policy=<type> : Specialization policy for intramodule calls 
+        --amalgamate=<file>        : Amalgamate the bitcode into a single <file> before linking (used to deal with duplicate symbols)
+        --intra-spec-policy=<type> : Specialization policy for intramodule calls
                                      (<type> should be either none, aggressive, nonrec-aggressive, bounded, or onlyonce)
-        --inter-spec-policy=<type> : Specialization policy for intermodule calls 
+        --inter-spec-policy=<type> : Specialization policy for intermodule calls
                                      (<type> should be either none, aggressive, nonrec-aggressive, bounded, or onlyonce)
         --max-bounded-spec=N       : Maximum number of function specialization if spec policy is bounded
+        --use-pointer-analysis     : Use pointer analysis for dealing with indirect calls
+        --use-seaopt               : Enable LLVM optimizer using seadsa-based alias analysis
+        --use-crabopt              : Enable LLVM optimizer using the Crab abstract interpreter
+        --enable-config-prime      : Enable dynamic analysis to propagate manifest data
         --disable-inlining         : Disable inlining
-        --force-inline-bounce      : Force inlining of bounce functions generated by devirt
         --force-inline-spec        : Force inlining of functions generated by specialization
-        --keep-external=<file>     : Pass a list of function names that should remain external.
-        --enable-config-prime      : Enable dynamic analysis to propagate manifest data (experimental)
-        --llpe                     : Use Smowton's LLPE for intra-module prunning (experimental)
+        --keep-external=<file>     : Pass a list of function names that should remain external
+        --entry-point              : Entry points of a library (function names separated by comma)
+        --remove-functions         : List of functions to be removed at the user's risk.
         --ipdse                    : Apply inter-procedural dead store elimination (experimental)
-        --mc-dce                   : Use model-checking to perform intra-module dead code elimination (experimental)
-        --ai-dce                   : Use invariants inferred by abstract interpretation for intra-module dce (experimental)
-        --amalgamate=<file>        : Amalgamate the bitcode into a single <file> before linking (used to deal with duplicate symbols)
+        --rop-guided-dce           : Use model-checking to remove functions with likely more ROP gadgets (experimental)
     """
 
 def entrypoint():
@@ -107,10 +104,10 @@ def entrypoint():
 
 
 def  usage(exe):
-    template = '{0} [--work-dir=<dir>]  [--force] [--help] [--stats] [--opt-stats] [--no-strip] [--verbose] [--debug-manager=] [--debug-pass=] [--debug] [--print-after-all] [--devirt=<type>] [--intra-spec-policy=<type>] [--inter-spec-policy=<type>] [--max-bounded-spec=N] [--disable-inlining] [--force-inline-bounce] [--force-inline-spec] [--keep-external=<file>] [--enable-config-prime] [--llpe] [--ipdse] [--mc-dce] [--ai-dce] <manifest>\n'
+    template = '{0} [--work-dir=<dir>]  [--force] [--help] [--stats] [--opt-stats] [--no-strip] [--verbose] [--debug-manager=] [--debug-pass=] [--debug] [--entry-point] [--print-after-all] [--intra-spec-policy=<type>] [--inter-spec-policy=<type>] [--max-bounded-spec=N] [--disable-inlining] [--use-pointer-analysis] [--use-seaopt] [--use-crabopt] [--force-inline-spec] [--keep-external=<file>] [--enable-config-prime] [--ipdse] [--rop-guided-dce] [--remove-functions] <manifest>\n'
     sys.stderr.write(template.format(exe))
 
-class Slash(object):
+class Slash:
 
     def __init__(self, argv):
         utils.setLogger()
@@ -119,17 +116,14 @@ class Slash(object):
             cmdflags = ['work-dir=',
                         'force',
                         'no-strip',
-                        'devirt=',
                         'debug',
                         'debug-manager=',
                         'debug-pass=',
                         'print-after-all',
                         'enable-config-prime',
-                        'llpe',
                         'help',
                         'ipdse',
-                        'mc-dce',
-                        'ai-dce',
+                        'rop-guided-dce',
                         'info',
                         'opt-stats',
                         'stats',
@@ -137,19 +131,23 @@ class Slash(object):
                         'inter-spec-policy=',
                         'max-bounded-spec=',
                         'disable-inlining',
-                        'force-inline-bounce',
+                        'use-pointer-analysis',
+                        'use-seaopt',
+                        'use-crabopt',
                         'force-inline-spec',
                         'tool=',
                         'verbose',
                         'keep-external=',
-                        'amalgamate=']
+                        'amalgamate=',
+                        'entry-point=',
+                        'remove-functions=']
             parsedargs = getopt.getopt(argv[1:], None, cmdflags)
             (self.flags, self.args) = parsedargs
 
             tool = utils.get_flag(self.flags, 'tool', None)
             if tool is not None:
                 tool = config.get_llvm_tool(tool)
-                print('tool = {0}'.format(tool))
+                print(f'tool = {tool}')
                 sys.exit(0)
 
             helpme = utils.get_flag(self.flags, 'help', None)
@@ -181,7 +179,6 @@ class Slash(object):
         self.valid = True
         self.amalgamation = utils.get_amalgamation(self.flags)
 
-
     def run(self):
         """Run the main algorithm described in the paper Automated Software
            Winnowing (SAC 2015) available here
@@ -196,31 +193,12 @@ class Slash(object):
 
         def check_spec_policy(policy):
             """ Supported policies: none, aggressive, nonrec-aggressive, bounded or onlyonce """
-
-            if policy <> 'none' and \
-               policy <> 'aggressive' and \
-               policy <> 'bounded' and \
-               policy <> 'onlyonce' and \
-               policy <> 'nonrec-aggressive':
+            if not policy in ('none', 'aggressive', 'bounded', 'onlyonce', 'nonrec-aggressive'):
                 sys.stderr.write('Error: unsupported specialization policy. ' + \
-                                 'Valid policies: none, aggressive, nonrec-aggressive, bounded, onlyonce')
+                                 'Valid policies: none, aggressive, nonrec-aggressive, ' + \
+                                 'bounded, onlyonce\n')
                 return False
-            else:
-                return True
-
-        def check_devirt_method(method):
-            """ Supported methods: none, dsa, sea_dsa, cha_dsa """
-
-            if method <> 'none' and \
-               method <> 'dsa' and \
-               method <> 'sea_dsa' and \
-               method <> 'cha_dsa':
-                sys.stderr.write('Error: unsupported devirtualization method. ' + \
-                                 'Valid methods: none, dsa, sea_dsa, cha_dsa\n')
-                return False
-            else:
-                return True
-
+            return True
 
         parsed = utils.check_manifest(self.manifest)
 
@@ -229,80 +207,39 @@ class Slash(object):
         if not valid:
             return 1
 
-        (valid, module, binary, libs, native_libs, ldflags, args, name, constraints) = parsed
-
+        (valid, module, binary, libs, native_libs, ldflags, \
+         static_args, name, dynamic_args, lib_spec, main_spec) = parsed
 
         if not self.driver_config():
             return 1
 
-        no_strip = utils.get_flag(self.flags, 'no-strip', None)
-
-        use_config_prime = utils.get_flag(self.flags, 'enable-config-prime', None)
-        if use_config_prime is not None:
-            use_config_prime = True
-        else:
-            use_config_prime = False
-        
-        use_llpe = utils.get_flag(self.flags, 'llpe', None)
-        if use_llpe is not None:
-            use_llpe = True
-        else:
-            use_llpe = False
-
-        use_ipdse = utils.get_flag(self.flags, 'ipdse', None)
-        if use_ipdse is not None:
-            use_ipdse = True
-        else:
-            use_ipdse = False
-
-        use_mc_dce = utils.get_flag(self.flags, 'mc-dce', None)
-        if use_mc_dce is not None:
-            use_mc_dce = True
-        else:
-            use_mc_dce = False
-
-        use_ai_dce = utils.get_flag(self.flags, 'ai-dce', None)
-        if use_ai_dce is not None:
-            use_ai_dce = True
-        else:
-            use_ai_dce = False
-
+        no_strip = utils.get_bool_flag(self.flags, 'no-strip')
+        use_config_prime = utils.get_bool_flag(self.flags, 'enable-config-prime')
+        use_ipdse = utils.get_bool_flag(self.flags, 'ipdse')
+        use_rop_guided_dce = utils.get_bool_flag(self.flags, 'rop-guided-dce')
         show_stats = utils.get_flag(self.flags, 'stats', None)
         info = utils.get_flag(self.flags, 'info', None)
         if info is not None:
             show_stats = True
-
+        entry_point = utils.get_flag(self.flags,'entry-point','none')
+        use_library_spec = False
+        if lib_spec != []:
+            use_library_spec = True
+        remove_functions = utils.get_flag(self.flags,'remove-functions','none')
         intra_spec_policy = utils.get_flag(self.flags, 'intra-spec-policy', 'none')
         if not check_spec_policy(intra_spec_policy):
             return 1
-
         inter_spec_policy = utils.get_flag(self.flags, 'inter-spec-policy', 'none')
         if not check_spec_policy(inter_spec_policy):
             return 1
-
         # Only used if intra_spec_policy or inter_spec_policy = bounded
         max_bounded_spec = utils.get_flag(self.flags, 'max-bounded-spec', None)
-        
-        devirt = utils.get_flag(self.flags, 'devirt', 'none')
-        if not check_devirt_method(devirt):
-            return 1
-
-        no_inlining = utils.get_flag(self.flags, 'disable-inlining', None)
-        inline_bounce = utils.get_flag(self.flags, 'force-inline-bounce', None)
-        if inline_bounce is not None:
-            inline_bounce = True
-        else:
-            inline_bounce = False
-        inline_spec = utils.get_flag(self.flags, 'force-inline-spec', None)        
-        if inline_spec is not None:
-            inline_spec = True
-        else:
-            inline_spec = False
-        
-        sys.stderr.write('\nslash working on {0} wrt {1} ...\n'.format(module, ' '.join(libs)))
-
+        no_inlining = utils.get_bool_flag(self.flags, 'disable-inlining')
+        use_seadsa = utils.get_bool_flag(self.flags, 'use-pointer-analysis')
+        use_seaopt = utils.get_bool_flag(self.flags, 'use-seaopt')
+        use_crabopt = utils.get_bool_flag(self.flags, 'use-crabopt')
+        inline_spec = utils.get_bool_flag(self.flags, 'force-inline-spec')
         native_lib_flags = []
-
         #this is simplistic. we are assuming they are (possibly)
         #relative paths, if we need to use a search path then this
         #will need to be beefed up.
@@ -315,7 +252,7 @@ class Slash(object):
                 new_native_libs.append(os.path.realpath(lib))
         native_libs = new_native_libs
 
-        files = utils.populate_work_dir(module, libs, self.work_dir)
+        files = utils.populate_work_dir(module, libs, lib_spec, main_spec, self.work_dir)
         os.chdir(self.work_dir)
 
         profile_maps, profile_map_titles = [], []
@@ -345,11 +282,13 @@ class Slash(object):
             for ll in map(lambda t: list(t),
                           zip(*map(lambda OrdDic: OrdDic.iteritems(), profile_maps))):
                 k, j = None, 0
-                assert (len(ll) == len(profile_map_titles))
+                assert len(ll) == len(profile_map_titles)
                 for (ki, vi) in ll:
                     ki = f(ki)
-                    if k is None: k = ki
-                    else: assert(k == ki)
+                    if k is None:
+                        k = ki
+                    else:
+                        assert k == ki
                     print_file(k, vi, profile_map_titles[j])
                     j += 1
 
@@ -360,36 +299,74 @@ class Slash(object):
         # in this case we just want to show the stats and exit
         if info is not None:
             print_profile_maps()
-            return
+            return 0
+
+        def get_external(m):
+            "Get external functions for the main module"
+            pre = m.get()
+            post = m.new('ext')
+            passes.get_external_functions(pre,post)
+
+        def liboccamize(m):
+            "Create dummy main which invokes certain functions only"
+            pre = m.get()
+            post = m.new('lo')
+            lib_entry_functions = ""
+            if use_library_spec :
+                external_function_files = glob.glob("external.functions.*")
+                external_functions_list = []
+                for external_file in external_function_files:
+                    with open(external_file,"r") as func_file:
+                        print("Reading from file:\t"+external_file)
+                        spec_dict = json.load(func_file)
+
+                        if spec_dict['functions'] != []:
+                            external_functions_list += spec_dict['functions']
+                #  Take union across external functions files
+                external_functions_list = list(set(external_functions_list))
+                lib_entry_functions = ",".join(external_functions_list)
+            else:
+                lib_entry_functions = entry_point
+
+            print("\tEntry Point= "+lib_entry_functions)
+            passes.lib_occamize(pre,post,lib_entry_functions)
+
+        def remove_main(m):
+            "If library specialization remove the main function"
+            pre = m.get()
+            post = m.new('rm')
+            passes.remove_main(pre,post)
+
+        def _remove_functions(m):
+            "Remove functions and add runtime check if function is executed"
+            pre = m.get()
+            post = m.new('rem')
+            passes.remove_functions(pre,post,remove_functions)
+
+        if use_library_spec:
+            pool.InParallel(get_external, files.values(), self.pool)
+
+        if entry_point != "none" or use_library_spec:
+            pool.InParallel(liboccamize, files.values(), self.pool)
+
+        if remove_functions != "none":
+            pool.InParallel(_remove_functions, files.values(), self.pool)
 
         ### 0. Lift deployment information into main's module
-        if args is not None:
-            sys.stderr.write('Input-user specialization using fix parameters: {0}\n'.format(args))            
-            main = files[module]
-            pre = main.get()
-            post = main.new('a')
-            passes.specialize_program_args(pre, post, args, 'arguments', name=name)
-        elif constraints:
-            sys.stderr.write('Input-user specialization using the constraints: {0}\n'.format(constraints))
-            main = files[module]
-            pre = main.get()
-            post = main.new('a')
-            passes.constrain_program_args(pre, post, constraints, 'constraints')
+
+        main = files[module]
+        pre = main.get()
+        post = main.new('a')
+        passes.specialize_program_args(pre, post, name, static_args, dynamic_args, \
+                                       'arguments')
 
         if use_config_prime:
-            ## NEW: apply configuration prime in main
-            if args is not None:
-                pre = main.get()
-                post = main.new('cp')
-                # args are already lowered in the bitcode
-                passes.config_prime(pre, post, list(), 0)
-            elif constraints:
-                (num_unknown_args, known_args) = constraints
-                pre = main.get()
-                post = main.new('cp')
-                # known_args are already lowered in the bitcode
-                passes.config_prime(pre, post, list(), num_unknown_args)
-            
+            ## Apply configuration prime in main
+            pre = main.get()
+            post = main.new('cp')
+            # static_args are already lowered in the bitcode
+            passes.config_prime(pre, post, list(), dynamic_args)
+
         # Create interface for main. We can never internalize main
         interface.writeInterface(interface.mainInterface(), 'main.iface')
 
@@ -400,39 +377,43 @@ class Slash(object):
                                             'iface')
         refs = dict([(k, mkvf(k)) for (k, _) in vals])
 
-        def _references((m, f)):
-            "Computing references"
+        def compute_interfaces(x):
+            "Computing interfaces"
+            (m, f) = x
             nm = refs[m].new()
-            passes.interface(f.get(), nm, [])
+            ## False here means that we don't use_seadsa
+            passes.interface(f.get(), nm, [], False)
 
-        pool.InParallel(_references, vals, self.pool)
+        pool.InParallel(compute_interfaces, vals, self.pool)
         ### 2. And internalize everything that we can
-        def _internalize((m, i)):
-            "Internalizing from interfaces"
+        def internalize(x):
+            "Internalizing wrt interfaces"
+            (m, i) = x
             pre = i.get()
             post = i.new('i')
-            ifaces = [refs[f].get() for f in refs.keys() if f != m] + ['main.iface']
+            ifaces = [refs[f].get() for f in list(refs.keys()) if f != m] + ['main.iface']
             passes.internalize(pre, post, ifaces, self.whitelist)
 
-        pool.InParallel(_internalize, vals, self.pool)
+        pool.InParallel(internalize, vals, self.pool)
 
         # Begin main loop
         iface_before_file = provenance.VersionedFile('interface_before', 'iface')
         iface_after_file = provenance.VersionedFile('interface_after', 'iface')
         progress = True
         rewrite_files = {}
-        for m, _ in files.iteritems():
+        for m, _ in files.items():
             base = utils.prevent_collisions(m[:m.rfind('.bc')])
             rewrite_files[m] = provenance.VersionedFile(base, 'rw')
 
         # Options passed to the optimizer (opt)
         opt_options = []
-        if no_inlining is not None:
+        if no_inlining is True:
             opt_options += ['-disable-inlining']
 
         utils.write_timestamp("Started global fixpoint ...")
         iteration = 0
         max_fixpoint_iterations = 10 ## make this user parameter
+
         while progress:
             iteration += 1
             if iteration > max_fixpoint_iterations:
@@ -450,51 +431,53 @@ class Slash(object):
                 post = m.new('p')
                 post_base = os.path.basename(post)
                 fn = 'previrt_%s-%s' % (pre_base, post_base)
-                print "\tModule: " + str(pre)
-                print "\tIntra-specialization policy={0}".format(intra_spec_policy)
+                print("\tModule: " + str(pre))
+                print("\tIntra-specialization policy={0}".format(intra_spec_policy))
                 if intra_spec_policy == 'bounded':
-                    print "\tMax number of copies={0}".format(max_bounded_spec)
+                    print("\tMax number of copies={0}".format(max_bounded_spec))
                 passes.peval(pre, post, \
                              opt_options, \
                              intra_spec_policy, \
                              max_bounded_spec, \
-                             devirt, \
-                             inline_bounce, inline_spec, \
-                             use_llpe, use_ipdse, use_ai_dce, \
+                             use_seaopt, \
+                             use_seadsa, \
+                             inline_spec, \
+                             use_ipdse, use_crabopt, \
                              log=open(fn, 'w'))
 
             pool.InParallel(intra, files.values(), self.pool)
 
             ### 4. Gather Inter-module interfaces
-            iface = passes.deep([x.get() for x in files.values()],
-                                ['main.iface'])
+            iface = passes.propagate_interfaces([x.get() for x in files.values()],
+                                                ['main.iface'], use_seadsa)
             interface.writeInterface(iface, iface_before_file.new())
 
             ### 5. Inter-specialize
-            if inter_spec_policy <> 'none':
+            if inter_spec_policy != 'none':
 
                 # Create specialized functions and rewrite specification
-                def inter_spec((nm, m)):
+                def inter_specialize(x):
                     "Inter-module module specialization"
+                    (nm, m) = x
                     pre = m.get()
                     post = m.new('s')
                     rw = rewrite_files[nm].new()
                     passes.specialize(pre, post, rw, [iface_before_file.get()],
                                       inter_spec_policy, max_bounded_spec)
 
-                print "\tInter-specialization policy={0}".format(inter_spec_policy)
-                if inter_spec == 'bounded':
-                    print "\tMax number of copies={0}".format(max_bounded_spec)
-                    
-                pool.InParallel(inter_spec, files.items(), self.pool)
+                print("\tInter-specialization policy={0}".format(inter_spec_policy))
+                if inter_spec_policy == 'bounded':
+                    print("\tMax number of copies={0}".format(max_bounded_spec))
+
+                pool.InParallel(inter_specialize, files.items(), self.pool)
 
                 # Rewrite
-                def inter_rewrite((nm, m)):
+                def inter_rewrite(t):
                     "Inter-module module rewriting"
+                    (nm, m) = t
                     pre = m.get()
                     post = m.new('r')
-                    rws = [rewrite_files[x].get() for x in files.keys()
-                           if x != nm]
+                    rws = [rewrite_files[x].get() for x in list(files.keys()) if x != nm]
                     out = [None]
                     retcode = passes.rewrite(pre, post, rws, output=out)
                     fn = 'rewrite_%s-%s' % (os.path.basename(pre),
@@ -505,37 +488,43 @@ class Slash(object):
                     return retcode
 
                 rws = pool.InParallel(inter_rewrite, files.items(), self.pool)
-                
+
                 progress = any(rws)
             else:
-                print "Skipped inter-module specialization"
+                print("Skipped inter-module specialization")
 
             # Aggressive internalization
-            pool.InParallel(_references, vals, self.pool)
-            pool.InParallel(_internalize, vals, self.pool)
+            ## XXX: not sure why this is needed but it makes some difference.
+            pool.InParallel(compute_interfaces, vals, self.pool)
+            pool.InParallel(internalize, vals, self.pool)
 
             ### 6. Sealing
-            
+
             # Compute the interfaces again after new specialized functions
-            iface = passes.deep([x.get() for x in files.values()], ['main.iface'])
+            iface = passes.propagate_interfaces([x.get() for x in files.values()],
+                                                ['main.iface'], use_seadsa)
             interface.writeInterface(iface, iface_after_file.new())
 
-            # internalize 
+            # internalize
             def sealing(m):
                 "Hides exported functions that are not referenced from outside the module"
                 pre = m.get()
                 post = m.new('h')
                 passes.internalize(pre, post, [iface_after_file.get()], self.whitelist)
-                
+
             pool.InParallel(sealing, files.values(), self.pool)
 
+
         utils.write_timestamp("Finished global fixpoint.")
 
+        if entry_point != "none" or use_library_spec:
+            pool.InParallel(remove_main, files.values(), self.pool)
+
         # Strip everything
         # XXX: we strip symbols after the whole specialization process
         # has finished.  Otherwise, things can go wrong if intra-module
         # specialization occurs with all symbols stripped.
-        if no_strip is None:
+        if no_strip is False:
             def _strip(m):
                 "Stripping symbols"
                 pre = m.get()
@@ -548,79 +537,88 @@ class Slash(object):
             add_profile_map('after specialization')
 
         def link(binary, files, libs, native_libs, native_lib_flags, ldflags):
-            final_libs = [files[x].get() for x in libs]
+            final_libs = [files[x].get() for x in libs] + [files[x].get() for x in lib_spec]
             final_module = files[module].get()
             linker_args = None
             link_cmd = None
 
             if self.amalgamation is None:
-                linker_args = final_libs + native_libs + native_lib_flags + ldflags
-                link_cmd = '\nclang++ {0} -o {1} {2}\n'.format(final_module, binary, ' '.join(linker_args))
+                linker_args =  native_libs + native_lib_flags + ldflags
+                link_cmd = '\nclang++ {0} {1} -o {2} {3}\n'.format(final_module, ' '.join(final_libs), \
+                                                                   binary, ' '.join(linker_args))
             else:
                 linker_args = native_libs + native_lib_flags + ldflags
-                link_cmd = '\nclang++ {0} -o {1} {2}\n'.format(self.amalgamation, binary, ' '.join(linker_args))
-                # need to amalgamate the bitcode PRIOR to linking (only needed when there are duplicate symbols)
-                amalargs = ['-override', final_module] + final_libs + ['-o', self.amalgamation]
+                link_cmd = '\nclang++ {0} -o {1} {2}\n'.format(self.amalgamation, \
+                                                               binary, ' '.join(linker_args))
+                # need to amalgamate the bitcode PRIOR to linking
+                # (only needed when there are duplicate symbols)
+                amalargs = ['-override', final_module]  + ['-o', self.amalgamation]
                 driver.run('llvm-link', amalargs)
                 final_module = self.amalgamation
 
             sys.stderr.write('\nLinking ...\n')
             sys.stderr.write(link_cmd)
             try:
-                driver.linker(final_module, binary, linker_args)
+                driver.linker(final_module, final_libs, binary, linker_args)
                 sys.stderr.write('\ndone.\n')
                 return True
             except Exception:
                 sys.stderr.write('\nFAILED. Modify the manifest to add libraries and/or linker flags.\n\n')
-                import traceback
                 traceback.print_exc()
                 return False
 
-        link_ok = link(binary, files, libs, native_libs, native_lib_flags, ldflags)
 
-        if use_mc_dce and link_ok:
-            def mc_dce((m, ropfile)):
-                '''
-                Pruning using model-checking-based dce guided by maximizing
-                the number of removed ROP gadgets
-                '''
-                pre = m.get()
-                post = m.new('mc_dce')
-                try:
-                    ## TODO: extract entries from the interfaces. Right
-                    ## now, we will only perform DCE for the
-                    ## program but not for libraries.
-                    entries = ['main']
+        if entry_point == "none" and main_spec == []:
+            link_ok = link(binary, files, libs, native_libs, native_lib_flags, ldflags)
+        else:
+            print("Lib OCCAMIZE invoked, no linking neccesary")
+            link_ok = False
 
-                    ## TODO: make user options
-                    benefit_threshold = 20
-                    cost_threshold = 3
-                    timeout = 120
-                    memlimit = 4096
-                    return passes.mc_dce(pre, entries, ropfile, post,
-                                              benefit_threshold, cost_threshold,
-                                              timeout, memlimit, opt_options)
-                except Exception:
-                    sys.stderr.write("Precise dce failed on " + str(pre))
-                    return False
+        if use_rop_guided_dce and link_ok:
+            sys.stderr.write('\n\nIf you really want to use \"--rop-guided-dce\" option ' + \
+                             ' email me at jorge.navas@sri.com\n\n')
+            # def rop_guided_dce((m, ropfile)):
+            #     '''
+            #     Pruning using model-checking-based dce guided by maximizing
+            #     the number of removed ROP gadgets
+            #     '''
+            #     pre = m.get()
+            #     post = m.new('rop_guided_dce')
+            #     try:
+            #         ## TODO: extract entries from the interfaces. Right
+            #         ## now, we will only perform DCE for the
+            #         ## program but not for libraries.
+            #         entries = ['main']
 
-            ropgadget_cmd = utils.get_ropgadget()
-            if ropgadget_cmd is not None:
-                binary = os.path.join(os.path.dirname(os.path.abspath(files[module].get())), \
-                                      binary)
-                ropfile = binary + '.ropgadget.txt'
-                ropgadget_args = ['--binary', binary, '--silent', '--fns2lines', ropfile]
-                driver.run(ropgadget_cmd, ropgadget_args)
+            #         ## TODO: make user options
+            #         benefit_threshold = 20
+            #         cost_threshold = 3
+            #         timeout = 120
+            #         memlimit = 4096
+            #         return passes.rop_guided_dce(pre, entries, ropfile, post,
+            #                                     benefit_threshold, cost_threshold,
+            #                                     timeout, memlimit, opt_options)
+            #     except Exception:
+            #         sys.stderr.write("Precise dce failed on " + str(pre))
+            #         return False
 
-                mc_dce_args = map(lambda m: (m,ropfile), files.values())
-                rws = pool.InParallel(mc_dce, mc_dce_args , self.pool)
-                progress = any(rws)
-                if progress:
-                    if show_stats is not None:
-                        add_profile_map('after model-checking-based dce')
-                link(binary, files, libs, native_libs, native_lib_flags, ldflags)
-            else:
-                sys.stderr.write("ropgadget not found. Aborting model-checking-based dce ...")
+            # ropgadget_cmd = utils.get_ropgadget()
+            # if ropgadget_cmd is not None:
+            #     binary = os.path.join(os.path.dirname(os.path.abspath(files[module].get())), \
+            #                           binary)
+            #     ropfile = binary + '.ropgadget.txt'
+            #     ropgadget_args = ['--binary', binary, '--silent', '--fns2lines', ropfile]
+            #     driver.run(ropgadget_cmd, ropgadget_args)
+
+            #     rop_guided_dce_args = map(lambda m: (m,ropfile), files.values())
+            #     rws = pool.InParallel(rop_guided_dce, rop_guided_dce_args , self.pool)
+            #     progress = any(rws)
+            #     if progress:
+            #         if show_stats is not None:
+            #             add_profile_map('after model-checking-based dce')
+            #     link(binary, files, libs, native_libs, native_lib_flags, ldflags)
+            # else:
+            #     sys.stderr.write("ropgadget not found. Aborting model-checking-based dce ...")
 
         # Make symlinks for the "final" versions
         for x in files.values():
@@ -638,7 +636,7 @@ class Slash(object):
                 """
                 base = os.path.basename(abspath)
                 res = base.split(os.extsep)
-                assert(len(res) > 1)
+                assert len(res) > 1
                 return res[0]
             print_profile_maps(_splitext)
 
